
#include "g3d.h"
#include "g3dutil.h"

/*
   Compute plane constants from three points.
   Plane normal is given as a cross-product /n/ = /p - q/ x /p - r/, 
   the remaining unit vectors are established with no user control. 
   */
int /* TRUE if data in *pl valid;
       otherwise some error occurred and the constants
       could not be evaluated */
G3D_plane_consts_3(G3D_plane_t *pl,   /* OUT plane constants set in the 
                                          structure */
                   G3D_vect_t *p,       /* point on the plane */
                   G3D_vect_t *q,       /* point on the plane */
                   G3D_vect_t *r        /* point on the plane */
                   )
{
  G3D_vect_t pq, pr;
  register G3D_vect_t *n = &pl->n;

  VECMAC_3_VEC(pq, =, (*p), -, (*q));
  VECMAC_3_VEC(pr, =, (*p), -, (*r));
  VECMAC_CROSS((*n), =, pq, x, pr);
  if (VECMAC_NORM ((*n)) > 0) {
    VECMAC_NORMALIZE((*n));
    pl->c = VECMAC_DOT((*n), (*p));
    
    { /* Now establish the vectors /u/ and /v/  */
      register double nx = fabs(n->x);
      register double ny = fabs(n->y);
      register double nz = fabs(n->z);
      G3D_loc_t E1 = {1, 0, 0};
      G3D_loc_t E2 = {0, 1, 0};
      G3D_loc_t E3 = {0, 0, 1};
      
      if (nx > nz || ny > nz) {
        VECMAC_CROSS(pl->u, =, E3, x, pl->n);
      } else {
        if (ny > nx || nz > nx) {
          VECMAC_CROSS(pl->u, =, E1, x, pl->n);
        } else {
          VECMAC_CROSS(pl->u, =, E2, x, pl->n);
        } 
      }
      VECMAC_NORMALIZE(pl->u);
      VECMAC_CROSS(pl->v, =, pl->n, x, pl->u);
    }
    return TRUE;
  } else {
    return FALSE;
  }
}

/*
   Compute plane constants from three points.
   Plane normal is given as a cross-product /n/ = /p - q/ x /p - r/, 
   and the local coordinate system /u/, /v/, /n/ is established as follows: 
   /u/ = /p - q/ / ||p - q||, /v/ = /n/ x /u/. 
   */
int/* TRUE if data in *pl valid;
       otherwise some error occurred and the constants
       could not be evaluated */
G3D_plane_consts_3_alt(G3D_plane_t *pl,   /* OUT plane constants set in the */
                         /* structure */
                         G3D_vect_t *p,       /* IN point on the plane */
                         G3D_vect_t *q,       /* IN point on the plane */
                         G3D_vect_t *r        /* IN point on the plane */
                         )
{
  G3D_vect_t pq, pr;
  register G3D_vect_t *n = &pl->n;

  VECMAC_3_VEC(pq, =, (*p), -, (*q));
  VECMAC_3_VEC(pr, =, (*p), -, (*r));
  VECMAC_CROSS((*n), =, pq, x, pr);
  if (VECMAC_NORM ((*n)) > 0) {
    VECMAC_NORMALIZE((*n));
    pl->c = VECMAC_DOT((*n), (*p));
    
    { /* Now establish the vectors /u/ and /v/  */
      pl->u = pq;
      VECMAC_NORMALIZE(pl->u);
      VECMAC_CROSS(pl->v, =, pl->n, x, pl->u);
    }
    return TRUE;
  } else {
    return FALSE;
  }
}



int
G3D_plane_consts_4(G3D_plane_t *pl,   /* OUT plane constants set in the */
                                  /* structure */
                   G3D_vect_t *p,       /* IN point on the plane */
                   G3D_vect_t *q,       /* IN point on the plane */
                   G3D_vect_t *r,       /* IN point on the plane */
                   G3D_vect_t *s        /* IN point on the plane */
                   )
{
  G3D_vect_t pq2, sr2, qr2, ps2, pq, pr, qp, rq, sr, rqp, srq, *tmp;
  register G3D_vect_t *n = &pl->n;
  VECMAC_SCALAR(rqp, =, 0);
  VECMAC_SCALAR(srq, =, 0);

  VECMAC_3_VEC(qp, =, (*q), -, (*p));
  VECMAC_3_VEC(rq, =, (*r), -, (*q));
  VECMAC_3_VEC(sr, =, (*s), -, (*r));
  VECMAC_CROSS(rqp, =, rq, x, qp);
  VECMAC_CROSS(srq, =, sr, x, rq);
  if (VECMAC_DOT(rqp, srq) < 0) {  /* swap r and s */
    tmp = s; s = r; r = tmp;
  }
  
  VECMAC_LERP(pq2, 0.5, (*p), (*q));
  VECMAC_LERP(sr2, 0.5, (*s), (*r));
  VECMAC_LERP(qr2, 0.5, (*q), (*r));
  VECMAC_LERP(ps2, 0.5, (*p), (*s));
  VECMAC_3_VEC(pq, =, qr2, -, ps2);
  VECMAC_3_VEC(pr, =, sr2, -, pq2);
  VECMAC_CROSS((*n), =, pq, x, pr);
  VECMAC_NORMALIZE((*n));
  if (VECMAC_NORM ((*n)) > 0) {
    pl->c = VECMAC_DOT((*n), (*p));
    { /* Now establish the vectors /u/ and /v/  */
      register double nx = fabs(n->x), ny = fabs(n->y), nz = fabs(n->z);
      G3D_loc_t E1 = {1, 0, 0};
      G3D_loc_t E2 = {0, 1, 0};
      G3D_loc_t E3 = {0, 0, 1};
      
      if (nx > nz || ny > nz) {
        VECMAC_CROSS(pl->u, =, E3, x, pl->n);
      } else {
        if (ny > nx || nz > nx) {
          VECMAC_CROSS(pl->u, =, E1, x, pl->n);
        } else {
          VECMAC_CROSS(pl->u, =, E2, x, pl->n);
        }
      }
      VECMAC_NORMALIZE(pl->u);
      VECMAC_CROSS(pl->v, =, pl->n, x, pl->u);
    }
    return TRUE;
  } else {
    return FALSE;
  }
}


int
G3D_plane_consts_p_and_n (G3D_plane_t *pl, /* OUT plane constants set in the */
                                  /* structure */
                   G3D_vect_t *p,       /* IN point on the plane */
                   G3D_vect_t *n       /* IN normal to the plane */
                   )
{
  pl->n = *n;
  if (VECMAC_NORM ((*n)) > 0) {
    VECMAC_NORMALIZE (pl->n);
    pl->c = VECMAC_DOT (pl->n, (*p));
    
    { /* Now establish the vectors /u/ and /v/  */
      register double nx = fabs(pl->n.x);
      register double ny = fabs(pl->n.y);
      register double nz = fabs(pl->n.z);
      G3D_loc_t E1 = {1, 0, 0};
      G3D_loc_t E2 = {0, 1, 0};
      G3D_loc_t E3 = {0, 0, 1};
      
      if (nx > nz || ny > nz) {
        VECMAC_CROSS (pl->u, =, E3, x, pl->n);
      } else {
        if (ny > nx || nz > nx) {
          VECMAC_CROSS (pl->u, =, E1, x, pl->n);
        } else {
          VECMAC_CROSS (pl->u, =, E2, x, pl->n);
        }
      }
      VECMAC_NORMALIZE (pl->u);
      VECMAC_CROSS (pl->v, =, pl->n, x, pl->u);
    }
    return TRUE;
  } else {
    return FALSE;
  }
}



void
G3D_intersect_plane_and_line(G3D_vect_t *s, /* OUT intersection point */
                             double *lambda, /* OUT parameter to the */
                                            /* intersection point */
                             G3D_bool_t *parallel, /* OUT */
                             G3D_loc_t *a, /* IN point on line, */
                                         /* x = a + lambda * AB */
                             G3D_vect_t *AB, /* IN vector along line */
                             G3D_plane_t *p, /* IN plane, given by the */
                                             /* implicit equation n . x = c; */
                                             /* The parameters of the plane */
                                             /* are obtained by */
                                             /* plane_consts() */
                             double eps, /* IN tolerance, n . AB < eps */
                                         /* means that line coincides with */
                                         /* the plane or is parallel */
                             double infty /* IN `infinite' number to */
                                          /* set lambda to if line coincides */
                                          /* with the plane */
                         )
{
  double n_ab;
#define is_zero(num) (fabs(num) <= eps)
#define lin_comb(r, p, pq, l)                   \
  (r)->x = (p)->x + (l) * (pq)->x,              \
  (r)->y = (p)->y + (l) * (pq)->y,              \
  (r)->z = (p)->z + (l) * (pq)->z

  n_ab = VECMAC_DOT(p->n, (*AB));
  if (is_zero(n_ab)) {
    /* Line coincides with the plane or is parallel */
    *parallel = TRUE;
    if (is_zero(VECMAC_DOT(p->n, (*a)) - p->c)) *lambda = 0;
    else                                        *lambda = infty;
  } else {
    *parallel = FALSE;
    *lambda   = (p->c - VECMAC_DOT(p->n, (*a))) / n_ab;
  }
  lin_comb(s, a, AB, (*lambda));
}



double
G3D_signed_dist_from_plane(G3D_plane_t *p, /* IN plane */
                           G3D_loc_t *pt /* IN given point */
                           )
{
  double c_bar;

  c_bar = VECMAC_DOT(p->n, (*pt));
  return (c_bar - p->c);
}


int
G3D_plane_intersects_segment(G3D_plane_t *p,
                             G3D_loc_t *pt0, G3D_loc_t *pt1)
{
  double c0 = VECMAC_DOT(p->n, (*pt0));
  double c1 = VECMAC_DOT(p->n, (*pt1));
  double c  = p->c;
  if      (c0 >= c) {
    if (c1 <= c)
      return TRUE;
  } else {
    if (c1 >= c)
      return TRUE;
  }
  return FALSE;
}




int
G3D_intersect_plane_and_segment(G3D_vect_t *s,  /* OUT intersection point */
                                double *lambda, /* OUT parameter to the */
                                                /* intersection point */
                                G3D_bool_t *parallel, /* OUT */
                                G3D_loc_t *a, /* IN point on line, */
                                              /* x = a + lambda * AB */
                                G3D_vect_t *AB, /* IN vector along line */
                                G3D_plane_t *p, /* IN plane, given by the */
                                             /* implicit equation n . x = c; */
                                             /* The parameters of the plane */
                                             /* are obtained by */
                                             /* plane_consts() */
                                double eps, /* IN tolerance, n . AB < eps */
                                         /* means that line coincides with */
                                         /* the plane or is parallel */
                                double infty /* IN `infinite' number to */
                                          /* set lambda to if line coincides */
                                          /* with the plane */
                                )
{
  double n_ab;
#define is_zero(num) (fabs(num) <= eps)
#define lin_comb(r, p, pq, l)                   \
  (r)->x = (p)->x + (l) * (pq)->x,              \
  (r)->y = (p)->y + (l) * (pq)->y,              \
  (r)->z = (p)->z + (l) * (pq)->z
  double l;

  n_ab = VECMAC_DOT(p->n, (*AB));
  if (is_zero(n_ab)) {
    /* Line coincides with the plane or is parallel */
    *parallel = TRUE;
    if (is_zero(VECMAC_DOT(p->n, (*a)) - p->c)) l = 0;
    else                                      l = infty;
  } else {
    *parallel = FALSE;
    l         = (p->c - VECMAC_DOT(p->n, (*a))) / n_ab;
  }

  *lambda = l;
  if (l >= 0 && l <= 1) {
    lin_comb(s, a, AB, l);
    return TRUE;
  } else
    return FALSE;
}


/*
   Project points into plane.
   The projection coordinates are relative to zeroth point on input.
 */
void
G3D_plane_project_points (G3D_plane_t *p,
                          G3D_vect_t *input[],
                          int npts,
                          double uc[], double vc[]
                          )
{
  int i;
  G3D_vect_t o = (*input[0]), d, u = p->u, v = p->v;
  
  for (i = 0; i < npts; i++) {
    VECMAC_3_VEC (d, =, (*input[i]), -, o);
    uc[i] = VECMAC_DOT (d, u);
    vc[i] = VECMAC_DOT (d, v);
  }
}
