
#include "g3d.h"
#include "g3dutil.h"


#ifndef min
#   define min(A,B) (((A) < (B)) ? (A) : (B))
#endif
#ifndef max
#   define max(A,B) (((A) > (B)) ? (A) : (B))
#endif
#ifndef min_of_3
#   define min_of_3(A,B,C)           \
          (((A) < (B))?              \
           (((C) < (A))? (C): (A)):  \
           (((C) < (B))? (C): (B))   \
           )
#   define max_of_3(A,B,C)           \
          (((A) > (B))?              \
           (((C) > (A))? (C): (A)):  \
           (((C) > (B))? (C): (B))   \
           )
#endif

#define GET_LOC(P, N, loc)                       \
  {                                              \
    double x_, y_, z_;                           \
    (P)->get_loc((P)->vertex[N], &x_, &y_, &z_); \
    (loc).x = x_;                                \
    (loc).y = y_;                                \
    (loc).z = z_;                                \
  }


G3D_poly_t *
G3D_new_poly(int num_vert,
             G3D_vertex_t *vertex,
             G3D_vertex_coords_func get_loc
             )
{
  G3D_poly_t *p;

  p = g3d_new(G3D_poly_t);
  CHECK_MALLOC(p, return NULL);

  p->num_vert = 0;
  p->vertex   = NULL;

  return G3D_modify_poly(p, num_vert, vertex, get_loc);
}


static int
quad_is_planar(G3D_poly_t *p, double warp_frac);


G3D_bool_t
G3D_poly_is_planar(G3D_poly_t *p, double warp_frac)
{
  if (p->num_vert == 3)
    return TRUE;
  else if (p->num_vert == 4)
    return quad_is_planar(p, warp_frac);
  else
    return FALSE;               /* RAW -- need to check polys with */
                                /* more vertices */
}


static int
quad_is_planar(G3D_poly_t *p, double warp_frac)
{
  double l;
  G3D_vect_t v, n, p0, p1, p2, p3;
  G3D_plane_t P;

  GET_LOC(p, 0, p0);
  GET_LOC(p, 1, p1);
  GET_LOC(p, 2, p2);
  GET_LOC(p, 3, p3);
  G3D_plane_consts_3(&P, &p0, &p1, &p2);
  v3_sub(&p3, &p1, &v);
  l = v3_length(&v);
  n = P.n;
  v3_normalize(&n);
  return (fabs(v3_dot(&n, &v)) > warp_frac * l);
}


void
G3D_poly_recompute_bbox(G3D_poly_t *p)
{
  G3D_loc_t pt;
  int i;
  G3D_box_t *bb = &p->bounding_box;

  GET_LOC(p, 0, pt);
  bb->lower = bb->upper = pt;
  for (i = 1; i < p->num_vert; i++) {
    GET_LOC(p, i, pt);
    bb->lower.x = min(bb->lower.x, pt.x);
    bb->lower.y = min(bb->lower.y, pt.y);
    bb->lower.z = min(bb->lower.z, pt.z);
    bb->upper.x = max(bb->upper.x, pt.x);
    bb->upper.y = max(bb->upper.y, pt.y);
    bb->upper.z = max(bb->upper.z, pt.z);
  }
}


void
G3D_poly_bbox(G3D_poly_t *p, G3D_box_t *bb)
{
  bb->lower = p->bounding_box.lower;
  bb->upper = p->bounding_box.upper;
}





#define QP(DIR)                                    \
   (0.25 * (  f->v[0]->loc.DIR + f->v[1]->loc.DIR  \
            + f->v[2]->loc.DIR + f->v[3]->loc.DIR))

/* /tol/ is the tolerance in the physical space.  It is used to compute */
/* /eps/.  /eps/ is used in the test (lambda >= eps && lambda <= 1-eps). */
/* If eps == 0, the test is strict. */
/* If eg. eps == 0.001, the test allows also segments whose end point */
/* which is in fact invisible from the first point (i.e. it is behind */
/* the poly), but close to the surface, to be recognized as not intersecting */
/* the poly. */
G3D_bool_t
G3D_segment_intersects_poly(G3D_segment_t *s, G3D_poly_t *p, double tol)
{
  G3D_vect_t ab, ip;
  G3D_bool_t result = FALSE;
  double lambda;
  G3D_quad_t q;
  G3D_tri_t t;
  double eps, ab_length, ctol;
  G3D_box_t segbb;
#define FPT(n) f->v[n]->loc
#define TPT(n) t.loc[n]
#define QPT(n) q.loc[n]
#define ASS(bb, p0, p1, dir)                            \
  if (p0->dir <= p1->dir) {                             \
    bb.lower.dir = p0->dir; bb.upper.dir = p1->dir;     \
  } else {                                              \
    bb.lower.dir = p1->dir; bb.upper.dir = p0->dir;     \
  }
#define SEGBB(seg, bb)                                     \
  {                                                        \
    register G3D_loc_t *p0 = &seg->start, *p1 = &seg->end; \
                                                           \
    ASS(bb, p0, p1, x);                                    \
    ASS(bb, p0, p1, y);                                    \
    ASS(bb, p0, p1, z);                                    \
  }
#define BOXES_OVERLAP(r1, r2)                                         \
         (   (r1.lower.x <= r2.upper.x) && (r1.upper.x >= r2.lower.x) \
          && (r1.lower.y <= r2.upper.y) && (r1.upper.y >= r2.lower.y) \
          && (r1.lower.z <= r2.upper.z) && (r1.upper.z >= r2.lower.z) \
          )
  
  /* (i)   Find intersection of the segment with the plane of the */
  /*       face.  If there is no intersection, return FALSE; otherwise */
  /* (ii)  Test whether the intersection plane is inside the face */
  /*       (in its plane). */

  SEGBB(s, segbb);
  if (!(BOXES_OVERLAP(segbb, p->bounding_box)))/* first check bounding boxes */
    return FALSE;
  
  v3_sub(&s->end, &s->start, &ab);
  ab_length = v3_length(&ab);
  eps       = tol / ab_length;             /* gives d_lambda */
  ctol      = tol / 10;

  if        (p->num_vert == 3) {
    GET_LOC(p, 0, t.loc[0]); GET_LOC(p, 1, t.loc[1]); GET_LOC(p, 2, t.loc[2]);
    result
      = G3D_intersect_tri_and_line(&ip, &lambda, &s->start, &ab, &t, ctol);
  } else if (p->num_vert == 4) {
    GET_LOC(p, 0, q.loc[0]); GET_LOC(p, 1, q.loc[1]);
    GET_LOC(p, 2, q.loc[2]); GET_LOC(p, 3, q.loc[3]);
    result
      = G3D_intersect_quad_and_line(&ip, &lambda, &s->start, &ab, &q, ctol);
  } else {
    RAW /* Need to check for the general case */
  }

  return (result && lambda >= eps && lambda <= 1-eps);
}


int
G3D_get_vert_loc(G3D_poly_t *p, int vert, G3D_loc_t *loc)
{
  GET_LOC(p, vert, *loc);
  return 1;
}





G3D_poly_t *
G3D_modify_poly(G3D_poly_t *p,
                int num_vert,
                G3D_vertex_t *vertex,
                G3D_vertex_coords_func get_loc
                )
{
  int i;
  
  if (p->vertex != NULL)
    if (p->num_vert < num_vert) { g3d_free(p->vertex); p->vertex = NULL; }
  if (p->vertex == NULL)
    p->vertex   = g3d_new_array(num_vert, G3D_vertex_t);
  CHECK_MALLOC(p->vertex, return NULL);
  
  p->num_vert = num_vert;
  FOR(i, p->num_vert) {
    p->vertex[i] = vertex[i];
  }
  
  if (get_loc == NULL) {
    RAISE_ERR(G3D_NO_GET_LOC_FUNC);
    return NULL;
  }
  p->get_loc = get_loc;
  
  p->planar = G3D_poly_is_planar(p, 0.0);
  G3D_poly_recompute_bbox(p);
  
  return p;
}

