
#include "g3d.h"
#include "g3dutil.h"

#ifndef min
#   define min(A,B) (((A) < (B)) ? (A) : (B))
#endif
#ifndef max
#   define max(A,B) (((A) > (B)) ? (A) : (B))
#endif
#ifndef min_of_3
#   define min_of_3(A,B,C)           \
          (((A) < (B))?              \
           (((C) < (A))? (C): (A)):  \
           (((C) < (B))? (C): (B))   \
           )
#   define max_of_3(A,B,C)           \
          (((A) > (B))?              \
           (((C) > (A))? (C): (A)):  \
           (((C) > (B))? (C): (B))   \
           )
#endif

#define QLOC(q, n) &q->loc[n]
#define ALL_QUAD_LOCS  QLOC(q, 0), QLOC(q, 1), QLOC(q, 2), QLOC(q, 3)


#define QLOCN(Nmat, quad, dir) (  Nmat[0] * quad->loc[0].dir    \
                                + Nmat[1] * quad->loc[1].dir    \
                                + Nmat[2] * quad->loc[2].dir    \
                                + Nmat[3] * quad->loc[3].dir)

#define ALL_QUAD_COORDS(q) q->loc[0].x, q->loc[0].y,    \
                           q->loc[1].x, q->loc[1].y,    \
                           q->loc[2].x, q->loc[2].y,    \
                           q->loc[3].x, q->loc[3].y


#define N_LOC(q, n, v)                  \
 {                                      \
   v.x = (  n[0] * q->loc[0].x          \
          + n[1] * q->loc[1].x          \
          + n[2] * q->loc[2].x          \
          + n[3] * q->loc[3].x);        \
   v.y = (  n[0] * q->loc[0].y          \
          + n[1] * q->loc[1].y          \
          + n[2] * q->loc[2].y          \
          + n[3] * q->loc[3].y);        \
 }                                      \
   


void
G3D_q4_N(/* OUT shape function value corresponding to */
         double n[4],
         double xi, double eta
         /* IN coordinates in the double-unit square */
         )
            
{
  double one_minus_xi = (1 - xi);
  double one_plus_xi  = (1 + xi);
  double one_minus_eta = (1 - eta);
  double one_plus_eta  = (1 + eta);
  
  n[0] = 0.25 * one_minus_xi * one_minus_eta;
  n[1] = 0.25 * one_plus_xi  * one_minus_eta;
  n[2] = 0.25 * one_plus_xi  * one_plus_eta;
  n[3] = 0.25 * one_minus_xi * one_plus_eta;
}

/* Returns the derivative of the interpolation matrix {N} by xi, */
/* evaluated at a Gauss point eta, xi. */
void
G3D_q4_N_xi(double n_xi[4], double xi, double eta)
{
  double       n1, n2;

  n1  =   (1. - eta) * 0.25;
  n2  =   (1. + eta) * 0.25;
  
  n_xi[0] = -n1;
  n_xi[1] =  n1;
  n_xi[2] =  n2;
  n_xi[3] = -n2;
}


/* Returns the derivative of the submatrix {N2} of the displacement */
/* interpolation matrix {N} by eta, evaluated at a Gauss point eta, xi. */
void
G3D_q4_N_eta(double n_eta[4], double xi, double eta)
{
  double       n1, n2;

  n1  =   (1. - xi) * 0.25;
  n2  =   (1. + xi) * 0.25;
  
  n_eta[0]  = -n1;
  n_eta[1]  = -n2;
  n_eta[2]  =  n2;
  n_eta[3]  =  n1;
}


/* Computes the jacobian matrix  J(x,y) */
void
G3D_q4_J(double J[2][2],            /* OUT */
         double xi, double eta,    /* IN */
         double x1, double y1,      /* IN */
         double x2, double y2,
         double x3, double y3,
         double x4, double y4
         ) 
{
  double N_xi[4];
  double N_eta[4];

  G3D_q4_N_xi(N_xi, xi, eta);
  G3D_q4_N_eta(N_eta, xi, eta);
  J[0][0] = (N_xi[0]*x1 + N_xi[1]*x2 + N_xi[2]*x3 + N_xi[3]*x4);
  J[0][1] = (N_xi[0]*y1 + N_xi[1]*y2 + N_xi[2]*y3 + N_xi[3]*y4);
  J[1][0] = (N_eta[0]*x1 + N_eta[1]*x2 + N_eta[2]*x3 + N_eta[3]*x4);
  J[1][1] = (N_eta[0]*y1 + N_eta[1]*y2 + N_eta[2]*y3 + N_eta[3]*y4);
}


double
G3D_q4_detJ(double J[2][2])
{
  return (J[0][0] * J[1][1] - J[0][1] * J[1][0]);
}



int
G3D_q4_invJ(double J[2][2], double inv_J[2][2])
{
  double det;

  det  = J[0][0] * J[1][1] - J[0][1] * J[1][0];
  if (det == 0) return 0;
  det  = 1 / det;
  inv_J[0][0] =  J[1][1] * det;
  inv_J[0][1] = -J[0][1] * det;
  inv_J[1][0] = -J[1][0] * det;
  inv_J[1][1] =  J[0][0] * det;
  return 1;
}



static G3D_vect_t p, t1, t2;
static double N[4];
static double N_xi[4];
static double N_eta[4];

int
G3D_q4_tangent_plane_consts(G3D_plane_t *P,
                               /* OUT plane constants set in the */
                               /* structure */
                            G3D_quad_t *q,    /* IN quad */
                            double xi, double eta /* IN parameters */
                            )
{
  G3D_q4_N(N, xi, eta);     /* point on the quad */
  p.x = QLOCN(N, q, x);
  p.y = QLOCN(N, q, y);
  p.z = QLOCN(N, q, z);
  G3D_q4_N_xi(N_xi, xi, eta); /* tangent vector along /xi/ */
  t1.x = QLOCN(N_xi, q, x);
  t1.y = QLOCN(N_xi, q, y);
  t1.z = QLOCN(N_xi, q, z);
  G3D_q4_N_eta(N_eta, xi, eta); /* tangent vector along /eta/ */
  t2.x = QLOCN(N_eta, q, x);
  t2.y = QLOCN(N_eta, q, y);
  t2.z = QLOCN(N_eta, q, z);
  
  if (v3_length(&t1) == 0)
    return 0; /* failure */
  if (v3_length(&t2) == 0)
    return 0; /* failure */

  /* normal */
  v3_cross(&t1, &t2, &P->n);
  v3_normalize(&P->n); 
  P->c = v3_dot(&P->n, &p);
  
  { /* Now establish the vectors /u/ and /v/  */
    register double nx = fabs(P->n.x), ny = fabs(P->n.y), nz = fabs(P->n.z);
    G3D_loc_t E1 = {1, 0, 0};
    G3D_loc_t E2 = {0, 1, 0};
    G3D_loc_t E3 = {0, 0, 1};
    
    if (nx > nz || ny > nz) {
      v3_cross(&E3, &P->n, &P->u);
    } else {
      if (ny > nx || nz > nx) {
        v3_cross(&E1, &P->n, &P->u);
      } else {
        v3_cross(&E2, &P->n, &P->u);
      }
    }
    v3_normalize(&P->u);
    v3_cross(&P->n, &P->u, &P->v);
  }
  return TRUE; /* success */
}


static double corner_xi[4] = {-1,  1,  1, -1};
static double corner_eta[4] = {-1, -1,  1,  1};

int
G3D_intersect_quad_and_line(G3D_loc_t *s, /* OUT intersection */
                            double *lambda, /* OUT line parameter */
                                                       /* determining /s/ */
                            G3D_loc_t *A, /* IN first point on line */
                            G3D_vect_t *AB, /* IN vector along line */
                            G3D_quad_t *q, /* IN quad */
                            double tol /* IN the point is computed with */
                                       /* the precision +/-tol in the */
                                       /* physical space. */
                            ) 
{
  G3D_plane_t tangent_plane;
  double Lam;
  G3D_bool_t parallel;
  G3D_vect_t xs, dp;
  double prevdist, dist, xi, eta, dxi, deta, a1, a2, alpha, b1, b2, over;
  int n_iter;
#define MAX_N_ITER 20
  int corner = -1;
  int iter_failure = FALSE;
  double eps = tol;

 retry:
  corner++;
  if (corner > 3) {
    if (iter_failure) {
      RAISE_ERR(G3D_NO_CONVERGENCE);
    }
    return FALSE;
  }

  iter_failure = FALSE;
  
  xi = corner_xi[corner]; eta = corner_eta[corner];
  if (G3D_q4_tangent_plane_consts(&tangent_plane, q, xi, eta)) {
    G3D_intersect_plane_and_line(&xs, &Lam, &parallel, A, AB, &tangent_plane,
                                 eps, G3D_INFTY);
    if (parallel) {
      goto retry;
    }
  }

  v3_sub(&xs, &p, &dp);
  prevdist = FLT_MAX;
  n_iter = 0;
  while ((dist = v3_length(&dp)) > tol) {
    b1   = v3_dot(&t1, &dp);
    b2   = v3_dot(&t2, &dp);
    a1   = v3_dot(&t1, &t1);
    a2   = v3_dot(&t2, &t2);
    alpha= v3_dot(&t1, &t2);
    over = (alpha * alpha - a1 * a2);
    if (over == 0) { /* This means co-linear vectors /t1/ and /t2/ */
      iter_failure = TRUE;
      goto retry;
    }
    over = 1 / over;
    dxi = (alpha * b2 - a2 * b1) * over;
    deta = (alpha * b1 - a1 * b2) * over;
    xi += dxi; eta += deta;
    if (!G3D_q4_tangent_plane_consts(&tangent_plane, q, xi, eta))
      goto retry;
    G3D_intersect_plane_and_line(&xs, &Lam, &parallel, A, AB, &tangent_plane,
                                 eps, G3D_INFTY);
    if (parallel) {
      goto retry;
    }
    if (prevdist < dist)
      goto retry;
    prevdist = dist;
    v3_sub(&xs, &p, &dp);
    if (n_iter++ > MAX_N_ITER) {
      iter_failure = TRUE; goto retry;
    }
  }
  if (   xi >= -1 && xi <= 1
      && eta >= -1 && eta <= 1) {
    *s      = xs;
    *lambda = Lam;
    return TRUE;
  } else
    return FALSE;
}

                               
int
G3D_is_inside_quad2d(double x, double y, /* point */
                     double x1, double y1, /* quad vertices */
                     double x2, double y2,
                     double x3, double y3,
                     double x4, double y4
                     )
{
  return (   G3D_is_inside_tri2d(x, y, x1, y1, x2, y2, x3, y3)
          || G3D_is_inside_tri2d(x, y, x1, y1, x3, y3, x4, y4));
}





int
G3D_quad_inv_map(G3D_quad2d_t *q, /* IN the hex */
                 G3D_loc2d_t *loc, /* IN the 2D location */
                 double eps,  /* IN tolerance for computation  */
                 G3D_loc2d_t *param_loc /* OUT the corresponding */
                 /* location in the parametric */
                 /* space */
                 )
{
  G3D_loc2d_t pl = {0, 0}, xbar, diff, dpl;
  double center[2];
  double dist0, dist, dist0eps;
  double J[2][2], inv_J[2][2], N[4];
#define QUAD_CENTER(q, center)                                          \
  {                                                                     \
    center[0] = q->loc[0].x + q->loc[1].x + q->loc[2].x + q->loc[3].x;  \
    center[1] = q->loc[0].y + q->loc[1].y + q->loc[2].y + q->loc[3].y;  \
    center[0] *= 0.25;                                                  \
    center[1] *= 0.25;                                                  \
  }

  QUAD_CENTER(q, center);
  xbar.x = center[0];
  xbar.y = center[1];
  dist0    = v2_length(v2_sub(loc, &xbar, &diff));
  dist     = dist0;
  dist0eps = max(dist0 * eps, 1.e-12);
  while ((dist = v2_length(v2_sub(loc, &xbar, &diff))) > dist0eps) {
    G3D_q4_J(J, pl.x, pl.y, ALL_QUAD_COORDS(q));
    G3D_q4_invJ(J, inv_J);/* RAW: no provision for singular J*/
    dpl.x = inv_J[0][0] * diff.x + inv_J[1][0] * diff.y;
    dpl.y = inv_J[0][1] * diff.x + inv_J[1][1] * diff.y;
    pl.x += dpl.x;
    pl.y += dpl.y;
    G3D_q4_N(N, pl.x, pl.y);
    N_LOC(q, N, xbar);
  }
  *param_loc = pl;
  return TRUE;
}



double
G3D_quad2d_area(double x1, double y1, /* quad vertices */
                double x2, double y2,
                double x3, double y3,
                double x4, double y4
                )
{
  return (   G3D_tri2d_area(x1, y1, x2, y2, x4, y4)
          +  G3D_tri2d_area(x2, y2, x3, y3, x4, y4));
}

void
G3D_q4_interp_opt (double n[4],
                   double n_x[4],
                   double n_y[4],
                   double *detJ,
                   double J[2][2],
                   double invJ[2][2],
                   double xi, double eta,
                   double x1, double y1, 
                   double x2, double y2, 
                   double x3, double y3, 
                   double x4, double y4)
{
  double n_xi[4], n_eta[4];
  double one_over_detJ;

  G3D_q4_N     (n, xi, eta);
  G3D_q4_N_xi  (n_xi, xi, eta);
  G3D_q4_N_eta (n_eta, xi, eta);
  G3D_q4_J (J, xi, eta, 
            x1, y1, x2,  y2, x3,  y3, x4,  y4);

  *detJ = G3D_q4_detJ (J);

  if (*detJ == 0) {
    RAISE_ERR(G3D_ZERO_JACOBIAN);
    return;
  }
  
  one_over_detJ  = 1 / *detJ;

#define COMPUTE_n_x_term(n_x, i, inv_J, n_xi, n_eta)   \
  n_x[i] = (  inv_J[0][0] * n_xi[i] + inv_J[0][1] * n_eta[i])
#define COMPUTE_n_x(n_x, inv_J, n_xi, n_eta)           \
  {                                                             \
    COMPUTE_n_x_term(n_x, 0, inv_J, n_xi, n_eta);      \
    COMPUTE_n_x_term(n_x, 1, inv_J, n_xi, n_eta);      \
    COMPUTE_n_x_term(n_x, 2, inv_J, n_xi, n_eta);      \
    COMPUTE_n_x_term(n_x, 3, inv_J, n_xi, n_eta);      \
  }
#define COMPUTE_n_y_term(n_y, i, inv_J, n_xi, n_eta)   \
  n_y[i] = (  inv_J[1][0] * n_xi[i] + inv_J[1][1] * n_eta[i])
#define COMPUTE_n_y(n_y, inv_J, n_xi, n_eta)           \
  {                                                             \
    COMPUTE_n_y_term(n_y, 0, inv_J, n_xi, n_eta);      \
    COMPUTE_n_y_term(n_y, 1, inv_J, n_xi, n_eta);      \
    COMPUTE_n_y_term(n_y, 2, inv_J, n_xi, n_eta);      \
    COMPUTE_n_y_term(n_y, 3, inv_J, n_xi, n_eta);      \
  }

  invJ[0][0] =  J[1][1] * one_over_detJ;
  invJ[0][1] = -J[0][1] * one_over_detJ;
  invJ[1][0] = -J[1][0] * one_over_detJ;
  invJ[1][1] =  J[0][0] * one_over_detJ;
  
  COMPUTE_n_x(n_x, invJ, n_xi, n_eta);
  COMPUTE_n_y(n_y, invJ, n_xi, n_eta);
}
