#include "g3d.h"
#include "g3dutil.h"

#ifndef min
#   define min(A,B) (((A) < (B)) ? (A) : (B))
#endif
#ifndef max
#   define max(A,B) (((A) > (B)) ? (A) : (B))
#endif
#ifndef min_of_3
#   define min_of_3(A,B,C)           \
          (((A) < (B))?              \
           (((C) < (A))? (C): (A)):  \
           (((C) < (B))? (C): (B))   \
           )
#   define max_of_3(A,B,C)           \
          (((A) > (B))?              \
           (((C) > (A))? (C): (A)):  \
           (((C) > (B))? (C): (B))   \
           )
#endif

static int 
ccw (double x0, double y0, double x1, double y1, double x2, double y2);
static int
is_inside_triangle (double x, double y, /* point */
                    double x1, double y1, /* triangle vertices */
                    double x2, double y2,
                    double x3, double y3);


static int
is_inside_triangle (double x, double y, /* point */
                    double x1, double y1, /* triangle vertices */
                    double x2, double y2,
                    double x3, double y3) 
{
  int orient0, orient1, orient2;

  orient0 = ccw (x1, y1, x2, y2, x, y);
  orient1 = ccw (x2, y2, x3, y3, x, y);
  orient2 = ccw (x3, y3, x1, y1, x, y);

  /* wildcard: go either way */
  if (orient0 == 0) 
    orient0 = orient1? orient1: orient2;
  if (orient1 == 0) 
    orient1 = orient0;
  if (orient2 == 0) 
    orient2 = orient0;

  return (orient0==orient1 && orient1==orient2);
} 


/*
	This routine determines if the progression of points p0->p1->p2
is counter-clockwise or not.  The return value is
	 1) counter-clockwise, 
	 0) if the the points are linear, and 
	-1) if they are clockwise.
Another way of interpreting this data is: 
	 1) point 2 is to the left of the line p0->p1, 
	 0) point 2 is on the line, and 
	-1) point 2 is to the right.
*/
static int 
ccw (double x0, double y0, double x1, double y1, double x2, double y2) 
{
  double dx1, dx2, dy1, dy2;
  double area1, area2;

  dx1= x1 - x0, 	dy1= y1 - y0;
  dx2= x2 - x0, 	dy2= y2 - y0;

  area1= dx1*dy2;
  area2= dy1*dx2;

  if (area1 > area2) 
	return (1);	/* counter-clockwise, left */

  if (area1 < area2) 
	return (-1);	/* clockwise, right */

  return (0);	/* linear */
} 



int
G3D_intersect_tri_and_line (G3D_loc_t *s, /* OUT intersection */
                            double *lambda, /* OUT line parameter */
                            /* determining /s/ */
                            G3D_loc_t *A, /* IN first point on line */
                            G3D_vect_t *AB, /* IN vector along line */
                            G3D_tri_t *t, /* IN triangle */
                            double eps /* IN used in */
                            /* G3D_intersect_plane_and_line () */
                            ) 
{
  G3D_plane_t plane;
  G3D_loc_t r;
  G3D_bool_t parallel;
  double x0, y0, x1, y1, x2, y2;
    
  G3D_plane_consts_3 (&plane, &t->loc[0], &t->loc[1], &t->loc[2]);
  G3D_intersect_plane_and_line (s, lambda, &parallel, A, AB, &plane,
                               eps, G3D_INFTY);
  if (parallel) 
    return 0;

  VECMAC_3_VEC (r, =, t->loc[0], -, (*s));
  x0 = VECMAC_DOT (r, plane.u); y0 = VECMAC_DOT (r, plane.v);
  VECMAC_3_VEC (r, =, t->loc[1], -, (*s));
  x1 = VECMAC_DOT (r, plane.u); y1 = VECMAC_DOT (r, plane.v);
  VECMAC_3_VEC (r, =, t->loc[2], -, (*s));
  x2 = VECMAC_DOT (r, plane.u); y2 = VECMAC_DOT (r, plane.v);

  return is_inside_triangle (0, 0, x0, y0, x1, y1, x2, y2);
} 


int
G3D_is_inside_tri2d (double x, double y, /* point */
                     double x1, double y1, /* triangle vertices */
                     double x2, double y2,
                     double x3, double y3) 
{
  return is_inside_triangle (x, y, x1, y1, x2, y2, x3, y3);
} 






int
G3D_inv_map_tri2d (double  *L1, double  *L2, double  *L3, /* OUT the */
                   /* area coordinates; If /x/,/y/ inside, the */
                   /* function returns TRUE; otherwise FALSE is returned */
                   double x, double y, /* IN point */
                   double x1, double y1, /* IN triangle vertices */
                   double x2, double y2,
                   double x3, double y3
                   ) 
{
  double two_A, a1, a2, a3, b1, b2, b3, c1, c2, c3, two_A_inv;
  
  a1        = x2 * y3 - x3 * y2;
  a2        = x3 * y1 - x1 * y3;
  a3        = x1 * y2 - x2 * y1;
  two_A     = a1 + a2 + a3;
  two_A_inv = 1 / two_A;
  b1        = y2 - y3;
  b2        = y3 - y1;
  b3        = y1 - y2;
  c1        = x3 - x2;
  c2        = x1 - x3;
  c3        = x2 - x1;
  *L1       = (a1 + b1 * x + c1 * y) * two_A_inv;
  *L2       = (a2 + b2 * x + c2 * y) * two_A_inv;
  *L3       = (a3 + b3 * x + c3 * y) * two_A_inv;
  return (   *L1 >= 0 && *L1 <= 1
          && *L2 >= 0 && *L2 <= 1
          && *L3 >= 0 && *L3 <= 1);
} 



double
G3D_interpolate_on_tri2d (double x, double y,
                          double x1, double y1,
                          double x2, double y2,
                          double x3, double y3,
                          double v1, double v2, double v3) 
{
    double L1, L2, L3, two_A, a1, a2, a3, b1, b2, b3, c1, c2, c3, two_A_inv;

    a1 = x2 * y3 - x3 * y2;
    a2 = x3 * y1 - x1 * y3;
    a3 = x1 * y2 - x2 * y1;
    two_A = a1 + a2 + a3;
    if (two_A == 0) return FLT_MAX;
    two_A_inv = 1 / two_A;
    b1 = y2 - y3;
    b2 = y3 - y1;
    b3 = y1 - y2;
    c1 = x3 - x2;
    c2 = x1 - x3;
    c3 = x2 - x1;
    L1 = (a1 + b1 * x + c1 * y) * two_A_inv;
    L2 = (a2 + b2 * x + c2 * y) * two_A_inv;
    L3 = (a3 + b3 * x + c3 * y) * two_A_inv;
    return (L1 * v1 + L2 * v2 + L3 * v3);
} 




int
G3D_tri_inv_map (double x, double y,
                 double x1, double y1,
                 double x2, double y2,
                 double x3, double y3,
                 double *L1, double *L2, double *L3) 
{
  double two_A, a1, a2, a3, b1, b2, b3, c1, c2, c3, two_A_inv;
  
  a1 = x2 * y3 - x3 * y2;
  a2 = x3 * y1 - x1 * y3;
  a3 = x1 * y2 - x2 * y1;
  two_A = a1 + a2 + a3;

  if (two_A == 0) {
    *L1 = *L2 = *L3 = 0;
    return FALSE;               /* fail */
  } 
  
  two_A_inv = 1 / two_A;
  b1 = y2 - y3;
  b2 = y3 - y1;
  b3 = y1 - y2;
  c1 = x3 - x2;
  c2 = x1 - x3;
  c3 = x2 - x1;

  *L1 = (a1 + b1 * x + c1 * y) * two_A_inv;
  *L2 = (a2 + b2 * x + c2 * y) * two_A_inv;
  *L3 = (a3 + b3 * x + c3 * y) * two_A_inv;

  return TRUE;
} 



int
G3D_tri_shp_funcs (double x, double y,
                   double x1, double y1,
                   double x2, double y2,
                   double x3, double y3,
                   double *L1, double *L2, double *L3,
                   double *L1_x, double *L2_x, double *L3_x,
                   double *L1_y, double *L2_y, double *L3_y) 
{
  double two_A, a1, a2, a3, b1, b2, b3, c1, c2, c3, two_A_inv;
  
  a1 = x2 * y3 - x3 * y2;
  a2 = x3 * y1 - x1 * y3;
  a3 = x1 * y2 - x2 * y1;
  two_A = a1 + a2 + a3;
  
  if (two_A == 0) {
    *L1 = *L2 = *L3 = 0;
    *L1_x = *L2_x = *L3_x = *L1_y = *L2_y = *L3_y = 0;
    return FALSE;               /* fail */
  } 
  
  two_A_inv = 1 / two_A;
  b1 = y2 - y3;
  b2 = y3 - y1;
  b3 = y1 - y2;
  c1 = x3 - x2;
  c2 = x1 - x3;
  c3 = x2 - x1;
  
  *L1 = (a1 + b1 * x + c1 * y) * two_A_inv;
  *L2 = (a2 + b2 * x + c2 * y) * two_A_inv;
  *L3 = (a3 + b3 * x + c3 * y) * two_A_inv;

  *L1_x = b1 * two_A_inv;
  *L2_x = b2 * two_A_inv;
  *L3_x = b3 * two_A_inv;
  
  *L1_y = c1 * two_A_inv;
  *L2_y = c2 * two_A_inv;
  *L3_y = c3 * two_A_inv;
  
  return TRUE;
} 

/*
  Compute shape function derivatives with respect to the spatial
  coordinates given the shape function values.
  Clearly, since the shape functions vary linearly, their derivatives
  are constant over the triangle, and we don't need to compute them
  from the location of the evaluation point.
 */
int
G3D_tri_shp_func_der (double x1, double y1,
                      double x2, double y2,
                      double x3, double y3,
                      double *L1_x, double *L2_x, double *L3_x,
                      double *L1_y, double *L2_y, double *L3_y) 
{
  double two_A, a1, a2, a3, b1, b2, b3, c1, c2, c3, two_A_inv;
  
  a1 = x2 * y3 - x3 * y2;
  a2 = x3 * y1 - x1 * y3;
  a3 = x1 * y2 - x2 * y1;
  two_A = a1 + a2 + a3;
  
  if (two_A == 0) {
    *L1_x = *L2_x = *L3_x = *L1_y = *L2_y = *L3_y = 0;
    return FALSE;               /* fail */
  } 
  
  two_A_inv = 1 / two_A;
  b1 = y2 - y3;
  b2 = y3 - y1;
  b3 = y1 - y2;
  c1 = x3 - x2;
  c2 = x1 - x3;
  c3 = x2 - x1;
  
  *L1_x = b1 * two_A_inv;
  *L2_x = b2 * two_A_inv;
  *L3_x = b3 * two_A_inv;
  
  *L1_y = c1 * two_A_inv;
  *L2_y = c2 * two_A_inv;
  *L3_y = c3 * two_A_inv;
  
  return TRUE;
} 


double
G3D_tri2d_area (double x1, double y1,
                double x2, double y2,
                double x3, double y3) 
{
    double a1, a2, a3;

    a1 = x2 * y3 - x3 * y2;
    a2 = x3 * y1 - x1 * y3;
    a3 = x1 * y2 - x2 * y1;
    return 0.5 * (a1 + a2 + a3);
} 


/* This method of triangle/segment intersection is implemented
   following the article by Christophe Schlick and Gilles Subrenat,
   "Ray Intersection of Tessellated Surfaces : Quadrangles versus Triangles"
   in Graphics Gems V (edited by A. Paeth), Academic Press (1994) .
   It is about 40% faster than the G3D_intersect_tri_and_line () function.
 */
int 
G3D_intersect_tri_and_seg (G3D_loc_t *s, /* OUT intersection */
                           double *lambda, /* OUT line parameter */
                           /* determining /s/ */
                           G3D_loc_t *P, /* IN first point on line */
                           G3D_vect_t *V, /* IN vector along line */
                           G3D_tri_t *t, /* IN triangle */
                           double eps /* IN used in */
                           /* G3D_intersect_plane_and_line () */
                           ) 
{
  G3D_vect_t AB, AC, N, AP, AM, AMAC =  {0,0,0} , ABAC =  {0,0,0} , ABAM =  {0,0,0} ;
  double nx, ny, nz, nn;
  int discard;
  double APN, VN, u = 0, v = 0, T, uv;
    
  VECMAC_3_VEC (AB, =, t->loc[1], -, t->loc[0]);
  VECMAC_3_VEC (AC, =, t->loc[2], -, t->loc[0]);
  VECMAC_CROSS (ABAC, =, AB, x, AC);
  nn = VECMAC_NORM (ABAC);
  if (nn == 0) 
    return 0; /* no normal -- colinear edges */

  nn = 1/nn;
  VECMAC_2_VEC (N, =, ABAC);
  VECMAC_SCALAR (N, *=, nn);     /* normalize */
  VECMAC_3_VEC (AP, =, (*P), -, t->loc[0]);
  APN = VECMAC_DOT (AP, N);
  VN  = VECMAC_DOT ((*V), N);
  
  if (VN == 0) 
    return 0; /* normal perpendicular to the vector
                 -- triangle & segment are parallel */
  
  T = - APN / VN;
  *lambda = T;

  if (T < 0 || T > 1) {
    return 0; /* intersection outside of segment */
  } 

  nx = fabs (N.x);
  ny = fabs (N.y);
  nz = fabs (N.z);
  if (nx > ny) 
    if (nx > nz) /* nx ny nz || nx nz ny */
      discard = 'x';
    else                        /* nz nx ny */
      discard = 'z';
  else
    if (ny > nz) /* ny nx nz || ny nz nx */
      discard = 'y';
    else
      discard = 'z';

  VECMAC_2_VEC ((*s), =, (*P));
 (*s) .x += T * (*V) .x;
 (*s) .y += T * (*V) .y;
 (*s) .z += T * (*V) .z;
  VECMAC_3_VEC (AM, =, (*s), -, t->loc[0]);
  VECMAC_CROSS (AMAC, =, AM, x, AC);
  VECMAC_CROSS (ABAM, =, AB, x, AM);
  switch (discard) {
  case 'x':
    u = AMAC.x / ABAC.x;
    v = ABAM.x / ABAC.x;
    break;
  case 'y':
    u = AMAC.y / ABAC.y;
    v = ABAM.y / ABAC.y;
    break;
  case 'z':
    u = AMAC.z / ABAC.z;
    v = ABAM.z / ABAC.z;
    break;  
  } 
  uv = u + v;
  
  return (( u >= 0 &&  u <= 1) &&
          ( v >= 0 &&  v <= 1) &&
          (uv >= 0 && uv <= 1));
} 



int 
G3D_pt_inside_tri (G3D_loc_t *s, /* IN point in the plane of the triangle */
                   G3D_tri_t *t, /* IN triangle */
                   G3D_plane_t *p /* IN triangle plane */
                   ) 
{
  G3D_vect_t AB, AC, AM, AMAC =  {0,0,0} , ABAC =  {0,0,0} , ABAM =  {0,0,0} ;
  int discard;
  double nx, ny, nz;
  double u = 0, v = 0, uv;
    
  VECMAC_3_VEC (AB, =, t->loc[1], -, t->loc[0]);
  VECMAC_3_VEC (AC, =, t->loc[2], -, t->loc[0]);
  VECMAC_CROSS (ABAC, =, AB, x, AC);

  nx = fabs (p->n.x);
  ny = fabs (p->n.y);
  nz = fabs (p->n.z);
  if (nx > ny) 
    if (nx > nz) /* nx ny nz || nx nz ny */
      discard = 'x';
    else                        /* nz nx ny */
      discard = 'z';
  else
    if (ny > nz) /* ny nx nz || ny nz nx */
      discard = 'y';
    else
      discard = 'z';

  VECMAC_3_VEC (AM, =, (*s), -, t->loc[0]);
  VECMAC_CROSS (AMAC, =, AM, x, AC);
  VECMAC_CROSS (ABAM, =, AB, x, AM);
  switch (discard) {
  case 'x':
    u = AMAC.x / ABAC.x;
    v = ABAM.x / ABAC.x;
    break;
  case 'y':
    u = AMAC.y / ABAC.y;
    v = ABAM.y / ABAC.y;
    break;
  case 'z':
  default:
    u = AMAC.z / ABAC.z;
    v = ABAM.z / ABAC.z;
    break;  
  } 
  uv = u + v;
  
  return (( u >= 0 &&  u <= 1) &&
          ( v >= 0 &&  v <= 1) &&
          (uv >= 0 && uv <= 1));
} 




/*
                                                                          
  Find the circumcenter of a 2D triangle.               
                                                                          
  The result is returned both in terms of x-y coordinates and xi-eta      
  coordinates.  The xi-eta coordinate system is defined in terms of the   
  triangle:  the origin of the triangle is the origin of the coordinate   
  system; the destination of the triangle is one unit along the xi axis;  
  and the apex of the triangle is one unit along the eta axis.            

  This code was adapted from the `triangle' code by
  Jonathan Richard Shewchuk,                                    
  School of Computer Science,                                    
  Carnegie Mellon University,                                    
  5000 Forbes Avenue,                                            
  Pittsburgh, Pennsylvania  15213-3891,                          
  jrs@cs.cmu.edu   
                                
*/

void 
G3D_circumcenter (double torg_x, double torg_y,
                  double tdest_x, double tdest_y,
                  double tapex_x,  double tapex_y,
                  double *circumcenter_x, double *circumcenter_y,
                  double *xi, double *eta
                  ) 
{
  double xdo, ydo, xao, yao;
  double dodist, aodist;
  double denominator;
  double dx, dy;

  /* Compute the circumcenter of the triangle. */
  xdo = tdest_x - torg_x;
  ydo = tdest_y - torg_y;
  xao = tapex_x - torg_x;
  yao = tapex_y - torg_y;
  dodist = xdo * xdo + ydo * ydo;
  aodist = xao * xao + yao * yao;

  denominator = 0.5 / (xdo * yao - xao * ydo);

  *circumcenter_x = torg_x - (ydo * aodist - yao * dodist) * denominator;  
  *circumcenter_y = torg_y + (xdo * aodist - xao * dodist) * denominator;  

  /* To interpolate point attributes for the new point inserted at  */
  /*   the circumcenter, define a coordinate system with a xi-axis, */
  /*   directed from the triangle's origin to its destination, and  */
  /*   an eta-axis, directed from its origin to its apex.           */
  /*   Calculate the xi and eta coordinates of the circumcenter.    */
  dx = *circumcenter_x - torg_x;
  dy = *circumcenter_y - torg_y;
  *xi = (dx * yao - xao * dy) * (2.0 * denominator);
  *eta = (xdo * dy - dx * ydo) * (2.0 * denominator);
} 



/*
   Compute circumsphere center of a 3-D triangle.
 */
int /* FALSE if the calculation failed for numerical reasons; TRUE otherwise */
G3D_tri_circumsphere (G3D_loc_t *pa, G3D_loc_t *pb, G3D_loc_t *pc,
                      G3D_loc_t *center
                      ) 
{ 
  GEP_MATRIX(a);
  double b[3];
  int p[3];
  G3D_loc_t BA, CA, N = {0,0,0};
  double AA;
  GEP_STACKINIT (a, 3);

  VECMAC_3_VEC (BA, =, (*pb), -, (*pa));
  VECMAC_3_VEC (CA, =, (*pc), -, (*pa));
  VECMAC_CROSS (N, =, BA, x, CA);

  GEP_ELEM (a, 0, 0) = BA.x;
  GEP_ELEM (a, 0, 1) = BA.y;
  GEP_ELEM (a, 0, 2) = BA.z;
  GEP_ELEM (a, 1, 0) = CA.x;
  GEP_ELEM (a, 1, 1) = CA.y;
  GEP_ELEM (a, 1, 2) = CA.z;
  GEP_ELEM (a, 2, 0) = N.x;
  GEP_ELEM (a, 2, 1) = N.y;
  GEP_ELEM (a, 2, 2) = N.z;

  AA = VECMAC_DOT ((*pa), (*pa));
  b[0] = 0.5 * (VECMAC_DOT ((*pb), (*pb)) - AA);
  b[1] = 0.5 * (VECMAC_DOT ((*pc), (*pc)) - AA);
  b[2] = VECMAC_DOT ((*pa), N);

  if (gep_sgefa (&a, p) != 0) {
    return FALSE;
  }
  gep_sgesl (&a, p, b, GEP_JOB_A_X_B);

  VECMAC_3_SCALAR ((*center), =, b[0], b[1], b[2]);
  
  return TRUE;
} 


static double 
line_length (double fx, double fy, double lx, double ly)
{
  double temp1, temp2;
  temp1 = (double)(fx - lx);  temp1 *= temp1;
  temp2 = (double)(fy - ly);  temp2 *= temp2;
  return sqrt(temp1 + temp2); 
}

/*
   Compute the triangle shape
   The following measures are used:
   o  sigma = maxside / inscribed_circle_radius
   o  tau   = maxside / minside
   max Sigma may be chosen as 6, and max Tau as about 3.
 */
void
G3D_triangle_shape (double x1, double y1,
		    double x2, double y2,
		    double x3, double y3,
                    double max_sigma,
                    double max_tau,
                    double *sigma, double *tau,
		    int *good, int *type)
{
  double sidea, sideb, sidec, temp1;
  double sidesum, area, inscribedr, circumscribedr, minside, maxside;

  sidea = line_length(x1, y1, x2, y2);
  sideb = line_length(x3, y3, x2, y2);
  sidec = line_length(x1, y1, x3, y3);
  sidesum = 0.5 * (sidea + sideb + sidec);
  area    = sidesum*(sidesum - sidea)*(sidesum - sideb)*(sidesum - sidec);
  area    = sqrt(area);
  minside = min (sidea, sideb); minside = min (minside, sidec);
  maxside = max (sidea, sideb); maxside = max (maxside, sidec);
  temp1   = sidea * sideb * sidec;
  
  circumscribedr = temp1 * 0.25 / (area + 0.00000001*temp1);
  inscribedr     = area / sidesum;
  *sigma         = maxside / (inscribedr + 0.00000001*maxside);
  *tau           = maxside / minside;

  if (*sigma > max_sigma) {
    *good = FALSE;
    if (*tau > max_tau)
      *type = G3D_TRI_ACUTE;
    else
      *type = G3D_TRI_OBTUSE;
  } else {
    *good = TRUE;
    *type = G3D_TRI_GOOD;
  }
}

/*
   Compute triad.
   Compute triad corresponding to G3D_tri_to_tri2d();
   The first vector points from pa to pb,
   the normal is given as cross product of (pb-pa)x(pc-pa).
 */
void
G3D_tri_to_tri2d_triad (G3D_loc_t *pa, G3D_loc_t *pb, G3D_loc_t *pc,
                        G3D_loc_t *f0, G3D_loc_t *f1, G3D_loc_t *f2)
{
  G3D_loc_t b, c, n={0,0,0};

  VECMAC_3_VEC (b, =, (*pb), -, (*pa)); /* first in-plane vector */
  VECMAC_NORMALIZE (b);

  VECMAC_3_VEC (c, =, (*pc), -, (*pa));

  VECMAC_CROSS (n, =, b, x, c); /* derive normal */
  VECMAC_NORMALIZE (n);

  VECMAC_CROSS (c, =, n, x, b); /* derive second in-plane vector */
  
  *f0 = b;
  *f1 = c;
  *f2 = n;
}
/*
   Convert to 2D.
   The point pa is converted to {0,0}, the point pb to {xb, 0}, and the point
   pc to {xc,yc}
 */
void
G3D_tri_to_tri2d (G3D_loc_t *pa, G3D_loc_t *pb, G3D_loc_t *pc,
                  double x[3], double y[3])
{
  G3D_loc_t b, c;
  double x2x1;

  x[0] = 0;
  y[0] = 0;
  VECMAC_3_VEC (b, = , (*pb), -, (*pa));
  x[1] = VECMAC_NORM (b);
  y[1] = 0;
  VECMAC_3_VEC (c, = , (*pc), -, (*pa));
  x[2] = VECMAC_DOT (b, c) / x[1];
  x2x1 = x[2] / x[1];
  VECMAC_2_VEC (c, -=, x2x1 * b);
  y[2] = VECMAC_NORM (c);
}

double
G3D_tri_area (G3D_loc_t *pa, G3D_loc_t *pb, G3D_loc_t *pc)
{
  double x[3], y[3];
  G3D_tri_to_tri2d (pa, pb, pc, x, y);
  return G3D_tri2d_area (x[0], y[0], x[1], y[1], x[2], y[2]);
}

/*
   Convert to 2D.
   See writeup for G3D_tri_to_tri2d() for locations of the
   triangle vertices.
 */
void
G3D_tri_2dpt (G3D_loc_t *pa, G3D_loc_t *pb, G3D_loc_t *pc,
              G3D_loc_t *pt,
              double *px, double *py)
{
  G3D_loc_t b, c, t;
  double bc;
  
  VECMAC_3_VEC (b, = , (*pb), -, (*pa));
  VECMAC_NORMALIZE (b);
  VECMAC_3_VEC (c, = , (*pc), -, (*pa));
  bc = VECMAC_DOT (b, c);
  VECMAC_2_VEC (c, -=, bc * b);
  VECMAC_NORMALIZE (c);
  VECMAC_3_VEC (t, = , (*pt), -, (*pa));
  *px = VECMAC_DOT (b, t);
  *py = VECMAC_DOT (c, t);
}
