
#define G3DTRI6_TOPO_PRIVATE_HEADER
#include "g3d.h"
#include "g3dutil.h"


#ifndef min
#   define min(A,B) (((A) < (B)) ? (A) : (B))
#endif
#ifndef max
#   define max(A,B) (((A) > (B)) ? (A) : (B))
#endif
#ifndef min_of_3
#   define min_of_3(A,B,C)           \
          (((A) < (B))?              \
           (((C) < (A))? (C): (A)):  \
           (((C) < (B))? (C): (B))   \
           )
#   define max_of_3(A,B,C)           \
          (((A) > (B))?              \
           (((C) > (A))? (C): (A)):  \
           (((C) > (B))? (C): (B))   \
           )
#endif

void
G3D_tri6_N (double n[6], double r, double s)
{
  double t = 1 - r - s;
  n[0] = t * (t + t - 1);
  n[1] = r * (r + r - 1);
  n[2] = s * (s + s - 1);
  n[3] = 4 * r * t;
  n[4] = 4 * r * s;
  n[5] = 4 * s * t;
}

void
G3D_tri6_N_r (double n_r[6], double r, double s)
{ /* [-3+4*r+4*s, 4*r-1, 0, 4-8*r-4*s, 4*s, -4*s] */
  double t = 1 - r - s;
#define COMPUTE_n_r(n_r, r, s, t)               \
  {                                             \
    n_r[0] = -3+4*r+4*s;                        \
    n_r[1] = 4*r-1;                             \
    n_r[2] = 0;                                 \
    n_r[3] =  4-8*r-4*s;                        \
    n_r[4] =  4*s;                              \
    n_r[5] = -4*s;                              \
  }
  COMPUTE_n_r (n_r, r, s, t);
}


void
G3D_tri6_N_s (double n_s[6], double r, double s)
{ /* [-3+4*r+4*s, 0, 4*s-1, -4*r, 4*r, 4-4*r-8*s] */
  double t = 1 - r - s;
#define COMPUTE_n_s(n_s, r, s, t)               \
  {                                             \
    n_s[0] = -3+4*r+4*s;                        \
    n_s[1] = 0;                                 \
    n_s[2] =  4*s-1;                            \
    n_s[3] =  -4*r;                             \
    n_s[4] =  4*r;                              \
    n_s[5] = 4-4*r-8*s;                         \
  }  
  COMPUTE_n_s (n_s, r, s, t); 
}


void
G3D_tri6_J ( /* OUT */
             double J[2][2],
             /* IN */
             double r, double s, 
             double x0, double y0, 
             double x1, double y1, 
             double x2, double y2, 
             double x3, double y3, 
             double x4, double y4, 
             double x5, double y5
             ) 
{
  double n_r[6], n_s[6];
  double t = 1 - r - s;

  COMPUTE_n_r (n_r, r, s, t);
  COMPUTE_n_s (n_s, r, s, t);

  /*
     The definition of the Jacobian matrix below is as follows:
     [J] =
     +                             +
     | x,xi      y,xi   |
     | x,eta     y,eta  |
     +                             +
     This is the TRANSPOSE of that defined
     by Hughes, The finite element method (p.119).
   */
#define COMPUTE_J(J, n_r, n_s,                                                                  \
                  x0,  y0,                                                                 \
                  x1,  y1,                                                                 \
                  x2,  y2,                                                                 \
                  x3,  y3,                                                                 \
                  x4,  y4,                                                                 \
                  x5,  y5)                                                                 \
  {                                                                                             \
    J[0][0] = (  x0*n_r[0] + x1*n_r[1] + x2*n_r[2] + x3*n_r[3] + x4*n_r[4] + x5*n_r[5]);        \
    J[0][1] = (  y0*n_r[0] + y1*n_r[1] + y2*n_r[2] + y3*n_r[3] + y4*n_r[4] + y5*n_r[5]);        \
                                                                                                \
    J[1][0] = (  x0*n_s[0] + x1*n_s[1] + x2*n_s[2] + x3*n_s[3] + x4*n_s[4] + x5*n_s[5]);        \
    J[1][1] = (  y0*n_s[0] + y1*n_s[1] + y2*n_s[2] + y3*n_s[3] + y4*n_s[4] + y5*n_s[5]);        \
  }

  COMPUTE_J (J, n_r, n_s,
             x0,  y0,  
             x1,  y1,  
             x2,  y2,  
             x3,  y3,  
             x4,  y4,  
             x5,  y5);
}


double
G3D_tri6_detJ (double J[2][2])
{

#define COMPUTE_detJ(J) (J[0][0] * J[1][1] - J[0][1] * J[1][0])

  return COMPUTE_detJ (J);
}



int
G3D_tri6_invJ (double J[2][2], double inv_J[2][2])  
{
  double one_over_detJ;

  one_over_detJ = COMPUTE_detJ (J);

  if (one_over_detJ == 0) {
    RAISE_ERR(G3D_ZERO_JACOBIAN);
    return FALSE;
  }
  
  one_over_detJ  = 1 / one_over_detJ;

#define COMPUTE_invJ(inv_J, J, one_over_detJ)   \
  {                                             \
    inv_J[0][0] =  J[1][1] * one_over_detJ;     \
    inv_J[0][1] = -J[0][1] * one_over_detJ;     \
    inv_J[1][0] = -J[1][0] * one_over_detJ;     \
    inv_J[1][1] =  J[0][0] * one_over_detJ;     \
  }

  COMPUTE_invJ (inv_J, J, one_over_detJ);
  
  return TRUE;
}





void
G3D_tri6_interp_opt (double n[6],
                      double n_x[6],
                      double n_y[6],
                      double *detJ,
                      double J[2][2],
                      double invJ[2][2],
                      double r, double s, 
                      double x0, double y0, 
                      double x1, double y1, 
                      double x2, double y2, 
                      double x3, double y3, 
                      double x4, double y4, 
                      double x5, double y5 
                      )
{
  double n_r[6], n_s[6];
  double t = 1 - r - s;
  
  COMPUTE_n_r (n_r, r, s, t);
  COMPUTE_n_s (n_s, r, s, t);

  COMPUTE_J (J, n_r, n_s, 
             x0,  y0,  
             x1,  y1,  
             x2,  y2,  
             x3,  y3,  
             x4,  y4,  
             x5,  y5)
  
  *detJ =   COMPUTE_detJ (J);
  
  if ((*detJ) <= 0) {
    RAISE_ERR(G3D_ZERO_JACOBIAN);
    return;
  }

  { double one_over_detj;
    one_over_detj = 1 / (*detJ);
    COMPUTE_invJ (invJ, J, one_over_detj);  
  }
  
  n_x[0] = n_r[0] * invJ[0][0] + n_s[0] * invJ[0][1];
  n_y[0] = n_r[0] * invJ[1][0] + n_s[0] * invJ[1][1];
  n_x[1] = n_r[1] * invJ[0][0] + n_s[1] * invJ[0][1];
  n_y[1] = n_r[1] * invJ[1][0] + n_s[1] * invJ[1][1];
  n_x[2] = n_r[2] * invJ[0][0] + n_s[2] * invJ[0][1];
  n_y[2] = n_r[2] * invJ[1][0] + n_s[2] * invJ[1][1];
  n_x[3] = n_r[3] * invJ[0][0] + n_s[3] * invJ[0][1];
  n_y[3] = n_r[3] * invJ[1][0] + n_s[3] * invJ[1][1];
  n_x[4] = n_r[4] * invJ[0][0] + n_s[4] * invJ[0][1];
  n_y[4] = n_r[4] * invJ[1][0] + n_s[4] * invJ[1][1];
  n_x[5] = n_r[5] * invJ[0][0] + n_s[5] * invJ[0][1];
  n_y[5] = n_r[5] * invJ[1][0] + n_s[5] * invJ[1][1];
}
